package client

import (
	"fmt"
	"net/http"

	"gitlab.bad-wolf.fr/wolf/hmac"
)

// SignerRoundTripper is a RoundTripper that will sign any request before
// performing it
type SignerRoundTripper interface {
	http.RoundTripper
}

type signerRoundTripper struct {
	signer              hmac.Signer
	key                 []byte
	signatureHeader     string
	defaultRoundTripper http.RoundTripper
}

// NewSignerRoundTripper return a new SignerRoundTripper.
// Params:
// 	signer: the hmac.Signer used to sign request
// 	signatureHeader: the name of the header which will contain the signature
func NewSignerRoundTripper(defaultRoundTripper http.RoundTripper, signer hmac.Signer, key []byte, signatureHeader string) http.RoundTripper {
	return &signerRoundTripper{
		signer:              signer,
		key:                 key,
		signatureHeader:     signatureHeader,
		defaultRoundTripper: defaultRoundTripper,
	}
}

func (rt *signerRoundTripper) RoundTrip(req *http.Request) (*http.Response, error) {
	signature, err := rt.signer.Sign(req, rt.key)
	if err != nil {
		return nil, fmt.Errorf("unable to generate signature: %w", err)
	}

	req.Header.Add(rt.signatureHeader, string(signature))

	return rt.defaultRoundTripper.RoundTrip(req)
}
