package requestencoder

import (
	"bytes"
	"crypto/md5"
	"crypto/sha256"
	"crypto/sha512"
	"encoding/base64"
	"fmt"
	"io"
	"io/ioutil"
	"net/http"
	"net/url"
	"regexp"
	"strings"
	"text/template"
)

type templateEncoderData struct {
	Req      *http.Request
	Body     []byte
	UserData interface{}
}

// TemplateEncoder encode a request base on the provided go template string
func TemplateEncoder(tmplt string, userData interface{}) (RequestEncoder, error) {

	// Create template
	encoderTemplate := template.New("encoderTemplate")

	// Add some functions for use in the template
	templateFuncs := make(template.FuncMap)
	templateFuncs["b64"] = tmpltFuncB64
	templateFuncs["sha256"] = tmpltFuncSha256
	templateFuncs["sha512"] = tmpltFuncSha512
	templateFuncs["md5"] = tmpltFuncMd5
	templateFuncs["in"] = tmpltFuncIn
	templateFuncs["inSlice"] = tmpltFuncInSlice
	templateFuncs["inEqualFold"] = tmpltFuncInEqualFold
	templateFuncs["inSliceEqualFold"] = tmpltFuncInSliceEqualFold
	templateFuncs["toLower"] = tmpltFuncToLower
	templateFuncs["toUpper"] = tmpltFuncToUpper
	templateFuncs["parseURL"] = tmpltFuncParseURL
	templateFuncs["parseQuery"] = tmpltFuncParseQuery
	templateFuncs["reReplace"] = tmpltFuncRegex
	templateFuncs["extractQueryString"] = tmpltFuncExtractQueryString
	templateFuncs["extractPath"] = tmpltFuncExtractPath

	// Parse user template
	encoderTemplate, err := encoderTemplate.
		Funcs(templateFuncs).
		Parse(tmplt)
	if err != nil {
		return nil, fmt.Errorf("unable to parse template: %w", err)
	}

	// Return the encoder
	return func(req *http.Request) ([]byte, error) {

		var bodyCopy []byte
		if req.Body != nil {
			bodyBufferCopy := bytes.NewBuffer(nil)
			tee := io.TeeReader(req.Body, bodyBufferCopy)
			bodyCopy, err = ioutil.ReadAll(tee)
			if err != nil {
				return nil, fmt.Errorf("unable to duplicated body: %w", err)
			}
			req.Body = io.NopCloser(bodyBufferCopy)
		}

		// The buffer receiving the encoded request data
		buf := bytes.NewBuffer(nil)
		if err := encoderTemplate.Execute(
			buf,
			templateEncoderData{
				Req:      req,
				UserData: userData,
				Body:     bodyCopy,
			},
		); err != nil {
			return nil, fmt.Errorf("unable to execute template: %w", err)
		}

		return buf.Bytes(), nil
	}, nil
}

// template functions

func tmpltFuncB64(d interface{}) (string, error) {
	var b []byte
	switch d := d.(type) {
	case []byte:
		b = d
	case string:
		b = []byte(d)
	default:
		return "", fmt.Errorf("wrong input type: %T", d)
	}
	return base64.StdEncoding.EncodeToString(b), nil
}

func tmpltFuncSha256(d interface{}) ([]byte, error) {
	var b []byte
	switch d := d.(type) {
	case []byte:
		b = d
	case string:
		b = []byte(d)
	default:
		return nil, fmt.Errorf("wrong input type: %T", d)
	}
	h := sha256.New()
	if _, err := h.Write(b); err != nil {
		return nil, err
	}
	return h.Sum(nil), nil
}

func tmpltFuncSha512(d interface{}) ([]byte, error) {
	var b []byte
	switch d := d.(type) {
	case []byte:
		b = d
	case string:
		b = []byte(d)
	default:
		return nil, fmt.Errorf("wrong input type: %T", d)
	}

	h := sha512.New()
	if _, err := h.Write(b); err != nil {
		return nil, err
	}
	return h.Sum(nil), nil
}

func tmpltFuncMd5(d interface{}) ([]byte, error) {
	var b []byte
	switch d := d.(type) {
	case []byte:
		b = d
	case string:
		b = []byte(d)
	default:
		return nil, fmt.Errorf("wrong input type: %T", d)
	}

	h := md5.New()
	if _, err := h.Write(b); err != nil {
		return nil, err
	}
	return h.Sum(nil), nil
}

func tmpltFuncInSlice(elem string, list []string) bool {
	for _, e := range list {
		if elem == e {
			return true
		}
	}
	return false
}

func tmpltFuncIn(elem string, list ...string) bool {
	return tmpltFuncInSlice(elem, list)
}

func tmpltFuncParseURL(s string) (string, error) {
	return url.PathUnescape(s)
}

func tmpltFuncParseQuery(s string) (string, error) {
	return url.QueryUnescape(s)
}

func tmpltFuncInSliceEqualFold(elem string, list []string) bool {
	for _, e := range list {
		if strings.EqualFold(elem, e) {
			return true
		}
	}
	return false
}

func tmpltFuncInEqualFold(elem string, list ...string) bool {
	return tmpltFuncInSliceEqualFold(elem, list)
}

func tmpltFuncToLower(s string) string {
	return strings.ToLower(s)
}

func tmpltFuncToUpper(s string) string {
	return strings.ToUpper(s)
}

func tmpltFuncRegex(data, regex, substitution string) (string, error) {
	re, err := regexp.Compile(regex)
	if err != nil {
		return "", fmt.Errorf("invalid regex: %w", err)
	}
	return re.ReplaceAllString(data, substitution), nil
}

func tmpltFuncExtractQueryString(data string) (string, error) {
	parsedURL, err := url.Parse(data)
	if err != nil {
		return "", err
	}
	return parsedURL.RawQuery, nil
}

func tmpltFuncExtractPath(data string) (string, error) {
	parsedURL, err := url.Parse(data)
	if err != nil {
		return "", err
	}
	return parsedURL.Path, nil
}
