package test

import (
	"bytes"
	"crypto/sha256"
	"encoding/base64"
	"encoding/json"
	"fmt"
	"io/ioutil"
	"net/http"
	"net/http/httptest"
	"testing"
	"time"

	"github.com/labstack/echo/v4"
	"gitlab.bad-wolf.fr/wolf/hmac"
	"gitlab.bad-wolf.fr/wolf/hmac/client"
	hmacMiddleware "gitlab.bad-wolf.fr/wolf/hmac/middleware/echo"
	"gitlab.bad-wolf.fr/wolf/hmac/requestencoder"
)

// Typical template example include:
// - url
// - method
// - body hash
// - every header expected the one containing the signature it-self not to signe
// 	 the signature
var signerTemplate = "host:{{.Req.Host}}\n" +
	"method:{{.Req.Method}}\n" +
	"path:{{.Req.URL.Path}}\n" +
	"rawQuery:{{.Req.URL.RawQuery}}\n" +
	`{{range $key, $value := .Req.Header -}}
		{{- if inSlice $key $.UserData.HeaderToSign -}}
			{{- range $value -}}
				{{- $key -}}={{- . -}};
			{{- end -}}
		{{- end -}}
	{{- end}}` +
	"body:{{md5 .Body | b64}}\n"

func serverTest(t *testing.T, signer hmac.Signer) *httptest.Server {
	e := echo.New()

	ksm := func(c echo.Context) (key []byte, signature []byte, err error) {
		signatureB64 := c.Request().Header.Get("X-Signature")
		return []byte("dummyKey"), []byte(signatureB64), err
	}

	unsecured := e.Group("/unsecured")
	unsecured.GET("/ping", func(c echo.Context) error {
		return c.JSON(
			http.StatusOK,
			map[string]string{
				"key": "value",
				"it":  "works",
			},
		)
	})

	auth := e.Group("/auth", hmacMiddleware.SignatureChecker(signer, ksm, true))
	auth.POST("/ping", func(c echo.Context) error {
		body, err := ioutil.ReadAll(c.Request().Body)
		if err != nil {
			return err
		}
		return c.JSON(
			http.StatusOK,
			map[string]string{
				"there's": "authentication",
				"body":    string(body),
			},
		)
	})
	authOpt := e.Group("/auth-opt", hmacMiddleware.SignatureChecker(signer, ksm, false))
	authOpt.POST("/ping", func(c echo.Context) error {
		body, err := ioutil.ReadAll(c.Request().Body)
		if err != nil {
			return err
		}
		return c.JSON(
			http.StatusOK,
			map[string]string{
				"auth": fmt.Sprintf("%t", hmacMiddleware.SignatureOK(c)),
				"body": string(body),
			},
		)
	})

	authOpt.POST("/mandatory/ping", func(c echo.Context) error {
		body, err := ioutil.ReadAll(c.Request().Body)
		if err != nil {
			return err
		}
		return c.JSON(
			http.StatusOK,
			map[string]string{
				"there's": "authentication",
				"body":    string(body),
			},
		)
	}, hmacMiddleware.SignatureEnforce())

	auth.GET("/mandatory/queryString", func(c echo.Context) error {
		queryParams := c.QueryParams().Encode()
		return c.JSON(
			http.StatusOK,
			map[string]string{
				"params": queryParams,
			},
		)
	})

	return httptest.NewServer(e.Server.Handler)
}

func clientTest(t *testing.T, endpoint string, signer hmac.Signer) {

	signerClient := http.Client{}
	signerClient.Transport = client.NewSignerRoundTripper(
		http.DefaultTransport,
		signer,
		[]byte("dummyKey"),
		"X-Signature",
	)
	t.Run("Test no authentication", func(t *testing.T) {
		testBody := bytes.NewBufferString("TEST_BODY")
		request, err := http.NewRequest(http.MethodGet, endpoint+"/unsecured/ping", testBody)
		if err != nil {
			t.Fatalf("error creating request: %s", err)
		}

		resp, err := http.DefaultClient.Do(request)
		if err != nil {
			t.Fatalf("error performing request: %s", err)
		}

		if resp.StatusCode != http.StatusOK {
			t.Fatalf("unexpected status code: %d", resp.StatusCode)
		}
	})

	t.Run("Test mandatory auth, not authenticated", func(t *testing.T) {
		testBody := bytes.NewBufferString("TEST_BODY")

		request, err := http.NewRequest(http.MethodPost, endpoint+"/auth/ping", testBody)
		if err != nil {
			t.Fatalf("error creating request: %s", err)
		}

		resp, err := http.DefaultClient.Do(request)
		if err != nil {
			t.Fatalf("error performing request: %s", err)
		}

		if resp.StatusCode != http.StatusUnauthorized {
			t.Fatalf("unexpected status code: %d", resp.StatusCode)
		}
	})

	t.Run("Test mandatory auth, authenticated", func(t *testing.T) {
		testBody := bytes.NewBufferString("TEST_BODY")

		request, err := http.NewRequest(http.MethodPost, endpoint+"/auth/ping", testBody)
		if err != nil {
			t.Fatalf("error creating request: %s", err)
		}

		resp, err := signerClient.Do(request)
		if err != nil {
			t.Fatalf("error performing request: %s", err)
		}

		body, _ := ioutil.ReadAll(resp.Body)
		if resp.StatusCode != http.StatusOK {
			t.Fatalf("unexpected status code: %d (body: %s", resp.StatusCode, body)
		}
	})

	t.Run("Test mandatory auth, authenticated, with headers", func(t *testing.T) {
		testBody := bytes.NewBufferString("TEST_BODY")

		request, err := http.NewRequest(http.MethodPost, endpoint+"/auth/ping", testBody)
		if err != nil {
			t.Fatalf("error creating request: %s", err)
		}

		request.Header.Add("x-user-id", "dummyUserId")
		request.Header.Add("x-session-id", "dummySessionId")
		request.Header.Add("x-timestamp", time.Now().String())

		resp, err := signerClient.Do(request)
		if err != nil {
			t.Fatalf("error performing request: %s", err)
		}

		body, _ := ioutil.ReadAll(resp.Body)
		if resp.StatusCode != http.StatusOK {
			t.Fatalf("unexpected status code: %d (body: %s", resp.StatusCode, body)
		}
	})

	t.Run("Test optionnal auth, unauthenticated", func(t *testing.T) {
		testBody := bytes.NewBufferString("TEST_BODY")

		request, err := http.NewRequest(http.MethodPost, endpoint+"/auth-opt/ping", testBody)
		if err != nil {
			t.Fatalf("error creating request: %s", err)
		}

		resp, err := http.DefaultClient.Do(request)
		if err != nil {
			t.Fatalf("error performing request: %s", err)
		}

		body, _ := ioutil.ReadAll(resp.Body)
		if resp.StatusCode != http.StatusOK {
			t.Fatalf("unexpected status code: %d (body: %s", resp.StatusCode, body)
		}

		bodyData := make(map[string]string)
		if err = json.Unmarshal(body, &bodyData); err != nil {
			t.Fatalf("error unmarshaling body: %s", err)
		}

		if bodyData["auth"] != "false" ||
			bodyData["userID"] != "" ||
			bodyData["body"] != "TEST_BODY" {
			t.Fatalf("wrong body: %s", body)
		}
	})

	t.Run("Test optionnal auth, authenticated", func(t *testing.T) {
		testBody := bytes.NewBufferString("TEST_BODY")

		request, err := http.NewRequest(http.MethodPost, endpoint+"/auth-opt/ping", testBody)
		if err != nil {
			t.Fatalf("error creating request: %s", err)
		}

		resp, err := signerClient.Do(request)
		if err != nil {
			t.Fatalf("error performing request: %s", err)
		}

		body, _ := ioutil.ReadAll(resp.Body)
		if resp.StatusCode != http.StatusOK {
			t.Fatalf("unexpected status code: %d (body: %s", resp.StatusCode, body)
		}

		bodyData := make(map[string]string)
		if err = json.Unmarshal(body, &bodyData); err != nil {
			t.Fatalf("error unmarshaling body: %s", err)
		}

		if bodyData["auth"] != "true" ||
			bodyData["body"] != "TEST_BODY" {
			t.Fatalf("wrong body: %s", body)

		}
	})

	t.Run("Test mandatory auth on optional, not authenticated", func(t *testing.T) {
		testBody := bytes.NewBufferString("TEST_BODY")

		request, err := http.NewRequest(http.MethodPost, endpoint+"/auth-opt/mandatory/ping", testBody)
		if err != nil {
			t.Fatalf("error creating request: %s", err)
		}

		resp, err := http.DefaultClient.Do(request)
		if err != nil {
			t.Fatalf("error performing request: %s", err)
		}

		body, _ := ioutil.ReadAll(resp.Body)
		if resp.StatusCode != http.StatusUnauthorized {
			t.Fatalf("unexpected status code: %d (body: %s", resp.StatusCode, body)
		}
	})

	t.Run("Test mandatory auth on optional, authenticated", func(t *testing.T) {
		testBody := bytes.NewBufferString("TEST_BODY")

		request, err := http.NewRequest(http.MethodPost, endpoint+"/auth-opt/mandatory/ping", testBody)
		if err != nil {
			t.Fatalf("error creating request: %s", err)
		}

		resp, err := signerClient.Do(request)
		if err != nil {
			t.Fatalf("error performing request: %s", err)
		}

		body, _ := ioutil.ReadAll(resp.Body)
		if resp.StatusCode != http.StatusOK {
			t.Fatalf("unexpected status code: %d (body: %s", resp.StatusCode, body)
		}
	})

	t.Run("Test mandatory auth with query string, authenticated", func(t *testing.T) {
		request, err := http.NewRequest(http.MethodGet, endpoint+"/auth/mandatory/queryString?firstKey=test&ids=1&ids=3", nil)
		if err != nil {
			t.Fatalf("error creating request: %s", err)
		}

		resp, err := signerClient.Do(request)
		if err != nil {
			t.Fatalf("error performing request: %s", err)
		}

		body, _ := ioutil.ReadAll(resp.Body)
		if resp.StatusCode != http.StatusOK {
			t.Fatalf("unexpected status code: %d (body: %s", resp.StatusCode, body)
		}

		bodyData := make(map[string]string)
		if err = json.Unmarshal(body, &bodyData); err != nil {
			t.Fatalf("error unmarshaling body: %s", err)
		}

		if bodyData["params"] != "firstKey=test&ids=1&ids=3" {
			t.Fatalf("wrong body: %s", body)

		}
	})
}

func TestHMAC(t *testing.T) {

	t.Run("Test template", func(t *testing.T) {

		templateEncoder, err := requestencoder.TemplateEncoder(
			signerTemplate,
			map[string]interface{}{
				"SignatureHeader": "X-Signature",
				"HeaderToSign": []string{
					"X-User-Id",
					"X-Session-Id",
					"X-Timestamp",
				},
			},
		)
		if err != nil {
			t.Errorf("unable to create templateEncoder: %s", err)
		}

		signer := hmac.NewSigner(
			hmac.SignerWithHashFunction(sha256.New),
			hmac.SignerWithRequestEncoder(templateEncoder),
		)

		req, err := http.NewRequest(
			http.MethodPost,
			"example.com/myRessource?arg=hello&arg2=world",
			bytes.NewBufferString("test body"),
		)

		if err != nil {
			t.Fatalf("unablte to create request: %s", err)
		}

		req.Header.Add("x-additional-header", "hello headers!")

		if _, err := signer.Sign(req, []byte("dummyKey")); err != nil {
			t.Fatalf("unablte to sign request: %s", err)
		}

	})

	t.Run("Test client/server", func(t *testing.T) {

		templateEncoder, err := requestencoder.TemplateEncoder(
			signerTemplate,
			map[string]interface{}{
				"SignatureHeader": "X-Signature",
				"HeaderToSign": []string{
					"X-User-Id",
					"X-Session-Id",
					"X-Timestamp",
				},
			},
		)
		if err != nil {
			t.Errorf("unable to create templateEncoder: %s", err)
		}

		signer := hmac.NewSigner(
			hmac.SignerWithHashFunction(sha256.New),
			hmac.SignerWithRequestEncoder(templateEncoder),
			hmac.SignerWithSignatureEncoder(
				func(b []byte) ([]byte, error) {
					return []byte(base64.StdEncoding.EncodeToString(b)), nil
				},
			),
		)

		testServer := serverTest(t, signer)
		defer testServer.Close()

		clientTest(t, testServer.URL, signer)
	})

}
